#include "haplotypetest.h"

/* Utility function */
extern void          *mem_alloc(size_t, const char *);
extern void           hpsort_reverse_index_of_double_array(int, double *, unsigned int *);
extern void           hpsort_index_of_double_array(int, double *, unsigned int *);

/* Error handler */
extern int            error_check();
extern int            error_message(char *);
extern int            warning_message(char *);

/* Statistic from P-value of chi square test */
extern double         calc_statistic_by_chi_square(int, double);

/* Odds ratio calculation for case-control study */
extern double         odds_ratio_by_penetrance(Estimate_Set *);

/* Define functions */
int            Confidence_interval_for_penetrance(Estimate_Set *, Option_Data *, double [][2], double ());
int            Confidence_interval_for_odds_ratio(Estimate_Set *, Option_Data *, double [2], double ());
int            Confidence_interval_for_QTL_parameters(Estimate_Set *, Genotype_Data *, Option_Data *, double [][2], double ());

int            Confidence_interval_for_penetrance(Estimate_Set *table,
						  Option_Data  *option,
						  double        q_bound[][2],
						  double        (*likelihood_function)())
{
  /* Local variables */
  double        conv  = 2.0 * log(10.0);
  int           RC, number_of_penetrance, i, c;
  double        p_value, x, L_max, L_0, L;
  double        q_save, q, q_min, q_max;
  double        q_range[2] = {0.0, 1.0};

  /* Get target L value */
  RC      = 0;
  p_value = 1.0 - option->confidence;
  x       = calc_statistic_by_chi_square(table->df_for_test, p_value);
  L_max   = (*likelihood_function)(table, table->f);
  L_0     = L_max - x/conv;
#ifdef DEBUG_CONFIDENCE
  printf("# %f %f %d %f %f\n",L_max,p_value,table->df_for_test,x,L_0);
#endif

  /* Get number_of penetrance */
  if (table->model_type == GENOTYPE_MODEL) {
    number_of_penetrance = 3;
  }
  else {
    number_of_penetrance = 2;
  }

  /* Loop for number of penetrance */
  for (i=0; i<number_of_penetrance; i++) {
    q_save = table->p_plus[i];

    /* Loop for lower and upper values */
    for (c=0; c<2; c++) {
      q_max  = q_save;
      q_min  = q_save;
      q      = (q_range[c] + q_max) * 0.5;
      /* Init loop */
      while (fabs(q - q_range[c]) > CRITERIA_CONF) {
	table->p_plus[i]  = q;
	table->p_minus[i] = 1.0 - q;
	L = (*likelihood_function)(table, table->f);
#ifdef DEBUG_CONFIDENCE
	printf("# init loop : %f %f %f %f %f\n",q,q_min,q_max,L,L_0);
#endif
	if (L < L_0) {
	  q_min = q;
	  break;
	}
	else {
	  q_max = q;
	}
	q = (q_range[c] + q) * 0.5;
      }

      /* Search loop */
      if (q_min != q_max) {
	q = (q_min + q_max) * 0.5;
	/*     while (fabs(L - L_0) > CRITERIA_CONF) {*/
	while (fabs(q_max - q_min) > CRITERIA_CONF) {
	  table->p_plus[i]  = q;
	  table->p_minus[i] = 1.0 - q;
	  L = (*likelihood_function)(table, table->f);
#ifdef DEBUG_CONFIDENCE
	  printf("# search loop : %f %f %f %f %f\n",q,q_min,q_max,L,L_0);
#endif
	  if (L > L_0) {
	    q_max = q;
	  }
	  else {
	    q_min = q;
	  }
	  q = (q_max + q_min) * 0.5;
	}
      }
      else {
	q = 0.0;
      }
      q_bound[i][c] = q;
#ifdef DEBUG_CONFIDENCE
      printf("# Limit value %d : %f %f %f\n",c,q,L,L_0);
#endif
    }

    table->p_plus[i]  = q_save;
    table->p_minus[i] = 1.0 - q_save;
  }

  return RC;
}

int            Confidence_interval_for_odds_ratio(Estimate_Set *table,
						  Option_Data  *option,
						  double        odds_bound[2],
						  double        (*likelihood_function)())
{
  /* Local variables */
  double        conv  = 2.0 * log(10.0);
  int           RC, l;
  double        p_value, L_max, L_0, L;
  double        q_save[2], a, b, c, d;
  double        dx_range[2], dx, dx_min, dx_max, rp, rm, odds_ratio;

  /* Get target L value */
  RC      = 0;
  p_value = 1.0 - option->confidence;
  dx      = calc_statistic_by_chi_square(table->df_for_test, p_value);
  L_max   = (*likelihood_function)(table, table->f);
  L_0     = L_max - dx/conv;
#ifdef DEBUG_CONFIDENCE
  printf("# %f %f %d %f %f\n",L_max,p_value,table->df_for_test,dx,L_0);
#endif

  /* Get number_of penetrance */
  if (table->model_type == GENOTYPE_MODEL) {
    odds_bound[0] = -1.0;
    odds_bound[1] = -1.0;
  }
  else {
    a = table->contingency[1][1]; /* a */
    b = table->contingency[1][0]; /* b */
    c = table->contingency[0][1]; /* c */
    d = table->contingency[0][0]; /* d */

    if ((a + b != 0) && (c + d != 0)) {

      q_save[0] = table->p_plus[0]; /* r- */
      q_save[1] = table->p_plus[1]; /* r+ */

      dx_range[0] = -a;
      if (dx_range[0] < -d) {
	dx_range[0] = -d;
      }
      dx_range[1] = b;
      if (dx_range[1] > c) {
	dx_range[1] = c;
      }
#ifdef DEBUG_CONFIDENCE
      printf("# table : %f %f\n",a,b);
      printf("#         %f %f\n",c,d);
      printf("# range : %f %f\n",dx_range[0],dx_range[1]);
#endif

      /* Loop for lower and upper values */
      for (l=0; l<2; l++) {
	dx_max  = 0.0;
	dx_min  = 0.0;
	dx      = (dx_range[l] + dx_max) * 0.5;
	/* Init loop */
	while (fabs(dx - dx_range[l]) > CRITERIA_CONF) {
	  rp                = (a + dx) / (a + b);
	  rm                = (c - dx) / (c + d);
	  table->p_plus[0]  = rm;
	  table->p_minus[0] = 1.0 - rm;
	  table->p_plus[1]  = rp;
	  table->p_minus[1] = 1.0 - rp;
	  L = (*likelihood_function)(table, table->f);
#ifdef DEBUG_CONFIDENCE
	  printf("# init loop : %f %f %f\n",dx,rp,rm);
	  printf("#           : %f %f %f %f\n",dx_min,dx_max,L,L_0);
#endif
	  if (L < L_0) {
	    dx_min = dx;
	    break;
	  }
	  else {
	    dx_max = dx;
	  }
	  dx = (dx_range[l] + dx) * 0.5;
	}

	/* Search loop */
	if (dx_min != dx_max) {
	  dx = (dx_min + dx_max) * 0.5;
	  /*     while (fabs(L - L_0) > CRITERIA_CONF) {*/
	  while (fabs(dx_max - dx_min) > CRITERIA_CONF) {
	    rp                = (a + dx) / (a + b);
	    rm                = (c - dx) / (c + d);
	    table->p_plus[0]  = rm;
	    table->p_minus[0] = 1.0 - rm;
	    table->p_plus[1]  = rp;
	    table->p_minus[1] = 1.0 - rp;
	    L = (*likelihood_function)(table, table->f);
#ifdef DEBUG_CONFIDENCE
	    printf("# search loop : %f %f %f\n",dx,rp,rm);
	    printf("#             : %f %f %f %f\n",dx_min,dx_max,L,L_0);
#endif
	    if (L > L_0) {
	      dx_max = dx;
	    }
	    else {
	      dx_min = dx;
	    }
	    dx = (dx_max + dx_min) * 0.5;
	  }
	  odds_ratio = odds_ratio_by_penetrance(table);
	}
	else {
	  odds_ratio = -1.0;
	}
	odds_bound[l] = odds_ratio;

#ifdef DEBUG_CONFIDENCE
	printf("# Limit value %d : %f %f %f %f\n",l,dx,odds_ratio,L,L_0);
#endif
      }

      table->contingency[1][1] = a;
      table->contingency[1][0] = b;
      table->contingency[0][1] = c;
      table->contingency[0][0] = d;
      table->p_plus[0]  = q_save[0];
      table->p_plus[1]  = q_save[1];
      table->p_minus[0] = 1.0 - q_save[0];
      table->p_minus[1] = 1.0 - q_save[1];
    }
    else {
      odds_bound[0] = -1.0;
      odds_bound[1] = -1.0;
    }
  }


  return RC;
}

int            Confidence_interval_for_QTL_parameters(Estimate_Set  *table,
						      Genotype_Data *genotype,
						      Option_Data   *option,
						      double         q_bound[][2],
						      double         (*likelihood_function)())
{
  /* Local variables */
  double        conv  = 2.0 * log(10.0);
  int           RC, number_of_mean, i, c;
  double        p_value, x, L_max, L_0, L;
  double        save, x_min, x_max, dx;

  /* Get target L value */
  RC      = 0;
  p_value = 1.0 - option->confidence;
  x       = calc_statistic_by_chi_square(table->df_for_test, p_value);
  L_max   = (*likelihood_function)(table, genotype, table->f);
  L_0     = L_max - x/conv;
#ifdef DEBUG_CONFIDENCE
  printf("# %f %f %d %f %f\n",L_max,p_value,table->df_for_test,x,L_0);
#endif

  /* Get number_of penetrance */
  if (table->model_type == GENOTYPE_MODEL) {
    number_of_mean = 3;
  }
  else {
    number_of_mean = 2;
  }

  /* Loop for number of mean */
  for (i=0; i<number_of_mean; i++) {
    save = table->mu[i];
    dx   = 0.2 * table->SD;

    /* Loop for lower and upper values */
    for (c=0; c<2; c++) {
      dx     = -dx;
      x_max  = save;
      x_min  = save + dx;
      x      = x_min;
      /* Init loop */
      while (1) {
	table->mu[i] = x;
	L = (*likelihood_function)(table, genotype, table->f);
#ifdef DEBUG_CONFIDENCE
	printf("# init loop : %f %f %f %f %f\n",x,x_min,x_max,L,L_0);
#endif
	if (L < L_0) {
	  break;
	}
	else {
	  x_max = x;
	  x_min = x + dx;
	}
	x = x_min;
      }

      /* Search loop */
      x = (x_min + x_max) * 0.5;
      /*     while (fabs(L - L_0) > CRITERIA_CONF) {*/
      while (fabs(x_max - x_min) > CRITERIA_CONF) {
	table->mu[i] = x;
	L = (*likelihood_function)(table, genotype, table->f);
#ifdef DEBUG_CONFIDENCE
	printf("# search loop : %f %f %f %f %f\n",x,x_min,x_max,L,L_0);
#endif
	if (L > L_0) {
	  x_max = x;
	}
	else {
	  x_min = x;
	}
	x = (x_max + x_min) * 0.5;
      }
      q_bound[i][c] = x;
#ifdef DEBUG_CONFIDENCE
      printf("# Limit value %d : %f %f %f\n",c,x,L,L_0);
#endif
    }

    table->mu[i] = save;
  }

  save = table->SD;
  dx   = 0.2 * table->SD;
  /* Loop for lower and upper values */
  for (c=0; c<2; c++) {
    dx     = -dx;
    x_max  = save;
    x_min  = save + dx;
    x      = x_min;
    /* Init loop */
    while (1) {
      table->SD = x;
      L = (*likelihood_function)(table, genotype, table->f);
#ifdef DEBUG_CONFIDENCE
      printf("# init loop : %f %f %f %f %f\n",x,x_min,x_max,L,L_0);
#endif
      if (L < L_0) {
	break;
      }
      else {
	x_max = x;
	x_min = x + dx;
      }
      x = x_min;
    }

    /* Search loop */
    x = (x_min + x_max) * 0.5;
    /*     while (fabs(L - L_0) > CRITERIA_CONF) {*/
    while (fabs(x_max - x_min) > CRITERIA_CONF) {
      table->SD = x;
      L = (*likelihood_function)(table, genotype, table->f);
#ifdef DEBUG_CONFIDENCE
      printf("# search loop : %f %f %f %f %f\n",x,x_min,x_max,L,L_0);
#endif
      if (L > L_0) {
	x_max = x;
      }
      else {
	x_min = x;
      }
      x = (x_max + x_min) * 0.5;
    }
    q_bound[number_of_mean][c] = x;
#ifdef DEBUG_CONFIDENCE
    printf("# Limit value %d : %f %f %f\n",c,x,L,L_0);
#endif
  }
  table->SD = save;

  return RC;
}

