#include "haplotypetest.h"

/* Utility function */
extern void          *mem_alloc(size_t, const char *);

/* Error handler */
extern int            error_check();
extern int            error_message(char *);
extern int            warning_message(char *);

/* Define functions */
Estimate_Set  *setup_EM(Genotype_Data *, Estimate_Set *, int);
int            pack_genotype(int, int *, Locus_Code *, Genotype_Pack *, int);
int            compare_genotype(int, Genotype_Pack *, Genotype_Pack *);

#ifdef DEBUG_PACK
#ifndef DEBUG
#define DEBUG
#endif
#endif

#ifdef DEBUG_COMBI
#ifndef DEBUG
#define DEBUG
#endif
#endif

#ifdef DEBUG_PENETRANCE
#ifndef DEBUG
#define DEBUG
#endif
#endif

#ifdef DEBUG_QTL
#ifndef DEBUG
#define DEBUG
#endif
#endif

#ifdef DEBUG_BOOTSTRAP_SAMPLING
#ifndef DEBUG
#define DEBUG
#endif
#endif

#ifdef DEBUG
char          *binary_string(haplotype);
int            dump_estimate_set(Estimate_Set *);
#endif

Estimate_Set  *setup_EM(Genotype_Data *genotype,
			Estimate_Set  *table,
			int            missing_criteria)
{
  /* Local variables */
  int             number_of_set;
  int             s, l, match, RC, reject, unknown;
  Genotype_Pack  *temp_geno_info;

  /* Memory allocation for the genotype packing table */
  table->geno_info = (Genotype_Pack *)mem_alloc(genotype->Number_of_Subjects*sizeof(Genotype_Pack),"Packed genotype structure table");
  temp_geno_info   = (Genotype_Pack *)mem_alloc(sizeof(Genotype_Pack),"Packed genotype sturucture");

  /* Pack genotype and merge packed genotype */
  number_of_set = 0;
  reject        = 0;
  unknown       = 0;
  for (s=0; s<genotype->Number_of_Subjects; s++) {
    /* Pack genotype to temporary variables */
    if (RC = pack_genotype(table->select_loci, table->loci, (genotype->Subjects + s)->Locus, temp_geno_info, missing_criteria)) {
      (genotype->Subjects + s)->packed_offset = -1;
      reject++;
      printf("Subject No %5d, %s (%d missing data loci) is not used for inference\n",s+1,(genotype->Subjects + s)->Name,RC);
    }
    else {
      /* Compare temporary variables and packed genotype array */
      temp_geno_info->phenotype = (genotype->Subjects + s)->phenotype;
      if (temp_geno_info->phenotype < 0) {
	unknown++;
      }
      if ((match = compare_genotype(number_of_set,table->geno_info,temp_geno_info)) != -1) {
	(genotype->Subjects + s)->packed_offset = match;
	(table->geno_info + match)->weight += 1.0;
      }
      else {
	(genotype->Subjects + s)->packed_offset = number_of_set;
	*(table->geno_info + number_of_set) = *temp_geno_info;
	number_of_set++;
      }
    }
  }

  free(temp_geno_info);
  genotype->Number_of_Effective_Subjects = genotype->Number_of_Subjects - reject;
  genotype->Number_of_Unknown            = unknown;
  table->number_of_set = number_of_set;
  table->normal        = 1.0/(double)genotype->Number_of_Effective_Subjects;
  table->normal_given  = 1.0/(double)(genotype->Number_of_Effective_Subjects - unknown);
  table->geno_info     = (Genotype_Pack *)realloc(table->geno_info,table->number_of_set*sizeof(Genotype_Pack));

  return table;
}

int            pack_genotype(int            number_of_loci,
			     int           *loci,
			     Locus_Code    *locus,
			     Genotype_Pack *geno_info,
			     int            missing_criteria)
{
  Locus_Code *l;
  int         i, RC, missing;
  allele      a1, a2;

  geno_info->Genotype_Code[0] = 0;
  geno_info->Genotype_Code[1] = 0;
  geno_info->Missing[0]       = 0;
  geno_info->Missing[1]       = 0;
  geno_info->weight           = 1.0;
  RC      = 0;
  missing = 0;
  for (i=0; i<number_of_loci; i++) {
    l  = locus + *(loci + i);
    a1 = ALLELE_CODE_MASK &  l->allele_code;
    a2 = ALLELE_CODE_MASK & (l->allele_code >> ALLELE_CODE_SHIFT);
    if (a1 != MISSING_DATA_CODE) {
      geno_info->Genotype_Code[0] = (geno_info->Genotype_Code[0] << 1) | a1;
    }
    else {
      geno_info->Genotype_Code[0] =  geno_info->Genotype_Code[0] << 1;
    }
    if (a2 != MISSING_DATA_CODE) {
      geno_info->Genotype_Code[1] = (geno_info->Genotype_Code[1] << 1) | a2;
    }
    else {
      geno_info->Genotype_Code[1] =  geno_info->Genotype_Code[1] << 1;
    }
    if      (a1 == MISSING_DATA_CODE) {
      geno_info->Missing[0]    = (geno_info->Missing[0] << 1) | 01;
      geno_info->Missing[1]    = (geno_info->Missing[1] << 1) | 01;
      missing++;
    }
    else if (a2 == MISSING_DATA_CODE) {
      geno_info->Missing[0]    = (geno_info->Missing[0] << 1) | 01;
      geno_info->Missing[1]    =  geno_info->Missing[1] << 1;
      missing++;
    }
    else {
      geno_info->Missing[0]    =  geno_info->Missing[0] << 1;
      geno_info->Missing[1]    =  geno_info->Missing[1] << 1;
    }
  }

  if (missing_criteria) {
    if (missing >= missing_criteria) {
      RC = missing;
    }
  }

#ifdef DEBUG_PACK
  printf("# packed = %32s ",
	 binary_string(geno_info->Genotype_Code[0]));
  printf("%32s\n",
	 binary_string(geno_info->Genotype_Code[1]));
  printf("           %32s ",
	 binary_string(geno_info->Missing[0]));
  printf("%32s\n",
	 binary_string(geno_info->Missing[1]));
#endif

  return RC;
}

#ifdef DEBUG
char          *binary_string(haplotype code)
{
  char *ptr;
  int   i, t;

  ptr = error_message_buffer;
  for (i=MAX_NUMBER_OF_LOCI-1; i>=0; i--) {
    t   = 01 & (code >> i);
    sprintf(ptr,"%1d",t);
    ptr++;
  }

  return error_message_buffer;
}

int            dump_estimate_set(Estimate_Set *table)
{
  int l;
  /* Debug output of data structure for EM */
  printf("# table->number_of_set  = %d\n",table->number_of_set);
  printf("# table->normal         = %f\n",table->normal);
  printf("# table->p_plus[0]      = %f\n",table->p_plus[0]);
  printf("# table->p_plus[1]      = %f\n",table->p_plus[1]);
  printf("# table->p_plus[2]      = %f\n",table->p_plus[2]);
  printf("# table->p_minus[0]     = %f\n",table->p_minus[0]);
  printf("# table->p_minus[1]     = %f\n",table->p_minus[1]);
  printf("# table->p_minus[2]     = %f\n",table->p_minus[2]);
  printf("# table->mu[0]          = %f\n",table->mu[0]);
  printf("# table->mu[1]          = %f\n",table->mu[1]);
  printf("# table->mu[2]          = %f\n",table->mu[2]);
  printf("# table->SD             = %f\n",table->SD);
  printf("# table->SD_c[0]        = %f\n",table->SD_c[0]);
  printf("# table->SD_c[1]        = %f\n",table->SD_c[1]);
  printf("# table->SD_c[2]        = %f\n",table->SD_c[2]);
  printf("# table->target         = %s\n",binary_string(table->target));
  printf("# table->target_mask    = %s\n",binary_string(table->target_mask));
  printf("# table->select_loci    = %d\n",table->select_loci);
  printf("# table->df             = %d\n",table->df);
  printf("# table->max_haplo_combi= %d\n",table->max_haplo_combi);
  printf("# table->number_of_haplo= %d\n",table->number_of_haplo);
  printf("# table->nonzero        = %d\n",table->nonzero);
  printf("# table->loci           =");
  for (l=0; l<table->select_loci; l++) {
    printf(" %d",*(table->loci + l));
  }
  printf("\n");
  printf("# table->f_allele       = %8x\n",table->f_allele);
  printf("# table->geno_info      = %8x\n",table->geno_info);
  printf("# table->combi          = %8x\n",table->combi);
  printf("# table->haplo          = %8x\n",table->haplo);
  printf("# table->class          = %8x\n",table->class);
  printf("# table->order          = %8x\n",table->order);
  printf("# table->f              = %8x\n",table->f);
  if (table->geno_info != NULL) {
    for (l=0; l<table->number_of_set; l++) {
      printf("# geno_info No. %5d\n",l);
      printf("  packed = %32s ",
	     binary_string((table->geno_info + l)->Genotype_Code[0]));
      printf("%32s\n",
	     binary_string((table->geno_info + l)->Genotype_Code[1]));
      printf("           %32s ",
	     binary_string((table->geno_info + l)->Missing[0]));
      printf("%32s\n",
	     binary_string((table->geno_info + l)->Missing[1]));
      printf("  phenotype = %d weight = %f\n",
	     (table->geno_info + l)->phenotype,
	     (table->geno_info + l)->weight);
    }
  }

  return 0;
}
#endif

int            compare_genotype(int            number_of_set,
				Genotype_Pack *pack_list,
				Genotype_Pack *pack_geno)
{
  int      i, match;

  match = -1;
  if (number_of_set > 0) {
    for (i=0; i<number_of_set; i++) {
      if (((pack_list+i)->Genotype_Code[0] == pack_geno->Genotype_Code[0]) &&
	  ((pack_list+i)->Genotype_Code[1] == pack_geno->Genotype_Code[1]) &&
	  ((pack_list+i)->Missing[0]       == pack_geno->Missing[0]      ) &&
	  ((pack_list+i)->Missing[1]       == pack_geno->Missing[1]      ) &&
	  ((pack_list+i)->phenotype        == pack_geno->phenotype       )) {
	match = i;
      }
    }
  }

#ifdef DEBUG_PACK
  printf("# number_of_set = %5d match = %5d\n",number_of_set,match);
#endif

  return match;
}
