#include "haplotypetest.h"

/* Utility function */
extern void          *mem_alloc(size_t, const char *);

/* Error handler */
extern int            error_check();
extern int            error_message(char *);
extern int            warning_message(char *);

/* Polymorphism parameters */
extern double         calc_PIC(int, double *);
extern double         calc_Heterozygosity(int, int, double *, double *);
extern double         calc_HWE_statistics(int *);
extern double         calc_p_value_by_chi_square(int, double);

/* Define functions */
Genotype_Data *get_genotype(char *, Option_Data *);
int            get_genotype_original(Genotype_Data *, Option_Data *);
int            get_genotype_prefile(Genotype_Data *, Option_Data *);
int            check_and_print_genotype(Genotype_Data *, Option_Data *);
int            set_allele_code(int, int, char *, Allele_Name *, int *, Locus_Code *, Locus_Data *);
int            check_comment(char *);
void           free_genotype(Genotype_Data *);
int            check_Subject_ID(Genotype_Data *, Option_Data *);

Genotype_Data *get_genotype(char *genotype_file, Option_Data *option)
{
  /* Local variables */
  char          *ext;
  Genotype_Data *genotype = NULL;

  /* Open genotype file */
  if (strlen(genotype_file) == 0) {
    error_message("Genotype file is not specified.");
    if (error_check()) return genotype;
  }
  else {
    File_Status->genotype_fp = fopen(genotype_file,"r");
  }

  if (File_Status->genotype_fp == NULL) {
    error_message("Genotype file open error.");
    if (error_check()) return genotype;
  }

  /* Copy genotype file name */
  strcpy(File_Status->genotype_name, genotype_file);
  ext = strstr(File_Status->genotype_name,".dat");
  if (ext != NULL) {
    *ext = '\0';
  }

  /* Memory allocation for genotype data */
  genotype = (Genotype_Data *)mem_alloc(sizeof(Genotype_Data),"Genotype data structure");
  genotype->Subjects   = NULL;
  genotype->Locus_List = NULL;

  switch (option->genotype_form) {
  case 0: /* original format */
    get_genotype_original(genotype, option);
    break;
  case 1: /* prefile format */
    get_genotype_prefile(genotype, option);
    break;
  default:
    error_message("Genotype file format cannot not be processed.");
    break;
  }

  fclose(File_Status->genotype_fp);
  File_Status->genotype_fp = NULL;

  if (error_check()) {
    free_genotype(genotype);
    genotype = NULL;
  }
  else {
    check_and_print_genotype(genotype,option);
  }

  return genotype;
}

int            get_genotype_original(Genotype_Data *genotype,
				     Option_Data   *option)
{
  /* Local variables for data processing */
  char             *read_buffer;
  int               not_found, size, subject_No;
  char             *ptr, *str, *end, *dlm;

  Subject_Genotype *subject;
  Locus_Code       *locus;
  int              *count;
  int               item_count, locus_No;
  Allele_Name      *allele_name;

  /* Intialize list size */
  genotype->Number_of_Subjects = 0;
  genotype->Number_of_Loci     = 0;
  read_buffer = (char *)mem_alloc(MAX_RECORD_LENGTH*sizeof(char),"Genotype file read buffer");

  /* Prescan genotype file and define size of genotype data */
  while(fgets(read_buffer, MAX_RECORD_LENGTH-1, File_Status->genotype_fp) != NULL) {
    /* Check record length */
    end = strchr(read_buffer, LINE_FEED);
    if (end == NULL) {
      error_message("Genotype file input buffer over flow. Change MAX_RECORD_LENGTH.");
      if (error_check()) {
	free(read_buffer); return 1;
      }
    }

    /* Delete \n */
    *end = STRING_TERMINATOR;

    /* Comment check and get string */
    if (!check_comment(read_buffer)) {
      ptr        = read_buffer;
      not_found  = 1;
      item_count = 0;

      while (*ptr != STRING_TERMINATOR) {
	/* If character is not delimiter, count only first character */
	if ((*ptr != DELIMITER_SPACE) && (*ptr != DELIMITER_TAB)) {
	  if (not_found) {
	    not_found = 0;
	    /* On first record, count number of loci */
	    if (genotype->Number_of_Subjects == 0) {
	      genotype->Number_of_Loci++;
	    }
	    /* Except first record, count number of loci */
	    else {
	      item_count++;
	    }
	  }
	}
	else {
	  not_found = 1;
	}
	ptr++;
      }

      if (genotype->Number_of_Subjects != 0) {
	if (genotype->Number_of_Loci != item_count) {
	  sprintf(error_message_buffer,"Number of genotype data for subject No. %d is differrent from the number of subject No. 1",genotype->Number_of_Subjects+1);
	  error_message(error_message_buffer);
	}
      }

      /* Count number of subjects */
      genotype->Number_of_Subjects++;
    }

  }
  /* Prescan of genotype file end */

  free(read_buffer);
  if (error_check()) {
     return 1;
  }

  /* In original_file format, first ORIGINAL_IGNORE_COUNT terms are not allele */
  genotype->Number_of_Loci = genotype->Number_of_Loci-ORIGINAL_IGNORE_COUNT;

#ifdef DEBUG
  printf("\n# Number of subjects = %d, number of loci =  %d\n",
	 genotype->Number_of_Subjects,genotype->Number_of_Loci);
#endif

  /* Memory allocation for subject and allele data */
  size  = genotype->Number_of_Subjects;
  genotype->Subjects   = (Subject_Genotype *)mem_alloc(size*sizeof(Subject_Genotype),"Subject array");
  size  = genotype->Number_of_Loci;
  genotype->Locus_List = (Locus_Data *)      mem_alloc(size*sizeof(Locus_Data),      "Locus data array");
  size  = genotype->Number_of_Subjects * genotype->Number_of_Loci;
  locus                = (Locus_Code *)      mem_alloc(size*sizeof(Locus_Code),      "Genotype array for each subject");

#ifdef DEBUG
  printf("# Array size for subject = %d, for locus =  %d, for genotype = %d\n",
	 genotype->Number_of_Subjects*sizeof(Subject_Genotype),
	 genotype->Number_of_Loci    *sizeof(Locus_Data),
	 size                        *sizeof(Locus_Code));
#endif

  /* Set address for allele data */
  subject = genotype->Subjects;
  for (subject_No=0; subject_No<genotype->Number_of_Subjects; subject_No++) {
    subject->Locus = locus;
    subject++; locus = locus + genotype->Number_of_Loci;
  }

  /* Initialize allele count */
  for (locus_No=0; locus_No<genotype->Number_of_Loci; locus_No++) {
    (genotype->Locus_List + locus_No)->count[0] = 0;
    (genotype->Locus_List + locus_No)->count[1] = 0;
    (genotype->Locus_List + locus_No)->count[2] = 0;
    (genotype->Locus_List + locus_No)->genotype_count[0] = 0;
    (genotype->Locus_List + locus_No)->genotype_count[1] = 0;
    (genotype->Locus_List + locus_No)->genotype_count[2] = 0;
  }

  /* Memory allocation for work area */
  read_buffer = (char *)mem_alloc(MAX_RECORD_LENGTH*sizeof(char),"Genotype file read buffer");
  size = genotype->Number_of_Loci;
  count       = (int *)        mem_alloc(size*sizeof(int),        "Work array to count allele");
  size = genotype->Number_of_Loci * MAX_INPUT_ALLELE;
  allele_name = (Allele_Name *)mem_alloc(size*sizeof(Allele_Name),"Work array to count allele");
  for (locus_No=0; locus_No<genotype->Number_of_Loci; locus_No++) {
    *(count+locus_No) = 0;
  }

  /* Process allele data */
  rewind(File_Status->genotype_fp);
  subject = genotype->Subjects;

  while(fgets(read_buffer, MAX_RECORD_LENGTH-1, File_Status->genotype_fp) != NULL) {
    /* Delete \n */
    end = strchr(read_buffer, LINE_FEED);
    *end = STRING_TERMINATOR;

    /* Comment check and get string */
    if (!check_comment(read_buffer)) {
      ptr        = read_buffer;
      not_found  = 1;
      item_count = 0;
      locus      = subject->Locus;
      locus_No   = 0;

      while (*ptr != STRING_TERMINATOR) {

	/* If character is not delimiter, count only first character */
	if ((*ptr != DELIMITER_SPACE) && (*ptr != DELIMITER_TAB)) {
	  if (not_found) {
	    not_found = 0;
	    str = ptr;
	  }
	}
	/* If character is delimiter, save allele name */
	else {
	  if (!not_found) {
	    if (item_count == 0) {
	      *ptr = STRING_TERMINATOR;
	      if (strlen(str) > CHAR_IN_SUBJECT_NAME) {
		sprintf(error_message_buffer,"Subject name, %s, is too long. First %d characters are used.",str,CHAR_IN_SUBJECT_NAME);
		warning_message(error_message_buffer);
		*(str+CHAR_IN_SUBJECT_NAME) = STRING_TERMINATOR;
	      }
	      strcpy(subject->Name,str);
	    }
	    else if (item_count == ORIGINAL_PHENOTYPE) {
	      *ptr = STRING_TERMINATOR;

	      if      (strcmp(str,"0") == 0) {
		if (option->switch_p) {
		  subject->phenotype = 1;
		}
		else {
		  subject->phenotype = 0;
		}
	      }
	      else if (strcmp(str,"1") == 0) {
		if (option->switch_p) {
		  subject->phenotype = 0;
		}
		else {
		  subject->phenotype = 1;
		}
	      }
	      else {
		subject->phenotype = -1;
	      }
	      subject->QTL       = 0.0;
	    }
	    else if (item_count >= ORIGINAL_IGNORE_COUNT) {
	      *ptr = STRING_TERMINATOR;

	      /* Find delimiter of alleles */
	      dlm = strchr(str, ORIGINAL_ALLELE_DELM);
	      if      (dlm == NULL) { /* Homozygote */
		set_allele_code(locus_No, 0, str, allele_name+MAX_INPUT_ALLELE*locus_No, count+locus_No, locus+locus_No, genotype->Locus_List+locus_No);
		set_allele_code(locus_No, 1, str, allele_name+MAX_INPUT_ALLELE*locus_No, count+locus_No, locus+locus_No, genotype->Locus_List+locus_No);
	      }
	      else {
		if (dlm == str) {       /* First allele is missing */
		  set_allele_code(locus_No, 0, MISSING_DATA_NAME, allele_name+MAX_INPUT_ALLELE*locus_No, count+locus_No, locus+locus_No, genotype->Locus_List+locus_No);
		}
		else {                  /* First allele */
		  *dlm = STRING_TERMINATOR;
		  set_allele_code(locus_No, 0, str, allele_name+MAX_INPUT_ALLELE*locus_No, count+locus_No, locus+locus_No, genotype->Locus_List+locus_No);
		}

		str = dlm+1;
		if (strlen(str) == 0) { /* Second allele is missing */
		  set_allele_code(locus_No, 1, MISSING_DATA_NAME, allele_name+MAX_INPUT_ALLELE*locus_No, count+locus_No, locus+locus_No, genotype->Locus_List+locus_No);
		}
		else {                  /* Second allele */
		  set_allele_code(locus_No, 1, str, allele_name+MAX_INPUT_ALLELE*locus_No, count+locus_No, locus+locus_No, genotype->Locus_List+locus_No);
		}
	      }

	      locus_No++;
	    }
	    item_count++;
	    not_found = 1;
	  }
	}

	ptr++;
      }

      /* Save last allele name */
      if (!not_found) {
	if (item_count >= ORIGINAL_IGNORE_COUNT) {
	  /* Find delimiter of alleles */
	  dlm = strchr(str, ORIGINAL_ALLELE_DELM);
	  if      (dlm == NULL) { /* Homozygote */
	    set_allele_code(locus_No, 0, str, allele_name+MAX_INPUT_ALLELE*locus_No, count+locus_No, locus+locus_No, genotype->Locus_List+locus_No);
	    set_allele_code(locus_No, 1, str, allele_name+MAX_INPUT_ALLELE*locus_No, count+locus_No, locus+locus_No, genotype->Locus_List+locus_No);
	  }
	  else {
	    if (dlm == str) {       /* First allele is missing */
	      set_allele_code(locus_No, 0, MISSING_DATA_NAME, allele_name+MAX_INPUT_ALLELE*locus_No, count+locus_No, locus+locus_No, genotype->Locus_List+locus_No);
	    }
	    else {                  /* First allele */
	      *dlm = STRING_TERMINATOR;
	      set_allele_code(locus_No, 0, str, allele_name+MAX_INPUT_ALLELE*locus_No, count+locus_No, locus+locus_No, genotype->Locus_List+locus_No);
	    }

	    str = dlm+1;
	    if (strlen(str) == 0) { /* Second allele is missing */
	      set_allele_code(locus_No, 1, MISSING_DATA_NAME, allele_name+MAX_INPUT_ALLELE*locus_No, count+locus_No, locus+locus_No, genotype->Locus_List+locus_No);
	    }
	    else {                  /* Second allele */
	      set_allele_code(locus_No, 1, str, allele_name+MAX_INPUT_ALLELE*locus_No, count+locus_No, locus+locus_No, genotype->Locus_List+locus_No);
	    }
	  }
	}
      }

      subject++;
    }
  }

  free(allele_name);
  free(count);
  free(read_buffer);

  return 0;
}

int            get_genotype_prefile(Genotype_Data *genotype,
				    Option_Data   *option)
{
  /* Local variables for data processing */
  char             *read_buffer;
  int               not_found, size, subject_No;
  char             *ptr, *str, *end;

  Subject_Genotype *subject;
  Locus_Code       *locus;
  int              *count;
  int               item_count, allele_No, locus_No;
  Allele_Name      *allele_name;

  /* Intialize list size */
  genotype->Number_of_Subjects = 0;
  genotype->Number_of_Loci     = 0;
  read_buffer = (char *)mem_alloc(MAX_RECORD_LENGTH*sizeof(char),"Genotype file read buffer");

  /* Prescan genotype file and define size of genotype data */
  while(fgets(read_buffer, MAX_RECORD_LENGTH-1, File_Status->genotype_fp) != NULL) {
    /* Check record length */
    end = strchr(read_buffer, LINE_FEED);
    if (end == NULL) {
      error_message("Genotype file input buffer over flow. Change MAX_RECORD_LENGTH.");
      if (error_check()) {
	free(read_buffer); return 1;
      }
    }

    /* Delete \n */
    *end = STRING_TERMINATOR;

    /* Comment check and get string */
    if (!check_comment(read_buffer)) {
      ptr        = read_buffer;
      not_found  = 1;
      item_count = 0;

      while (*ptr != STRING_TERMINATOR) {
	/* If character is not delimiter, count only first character */
	if ((*ptr != DELIMITER_SPACE) && (*ptr != DELIMITER_TAB)) {
	  if (not_found) {
	    not_found = 0;
	    /* On first record, count number of loci */
	    if (genotype->Number_of_Subjects == 0) {
	      genotype->Number_of_Loci++;
	    }
	    /* Except first record, count number of loci */
	    else {
	      item_count++;
	    }
	  }
	}
	else {
	  not_found = 1;
	}
	ptr++;
      }

      if (genotype->Number_of_Subjects != 0) {
	if (genotype->Number_of_Loci != item_count) {
	  sprintf(error_message_buffer,"Number of genotype data for subject No. %d is differrent from the number of subject No. 1",genotype->Number_of_Subjects+1);
	  error_message(error_message_buffer);
	}
      }

      /* Count number of subjects */
      genotype->Number_of_Subjects++;
    }

  }
  /* Prescan of genotype file end */

  free(read_buffer);
  if (error_check()) {
     return 1;
  }

  /* In prefile format, first PREFILE_IGNORE_COUNT terms are not allele */
  genotype->Number_of_Loci = (genotype->Number_of_Loci-PREFILE_IGNORE_COUNT)/2;

#ifdef DEBUG
  printf("\n# Number of subjects = %d, number of loci =  %d\n",
	 genotype->Number_of_Subjects,genotype->Number_of_Loci);
#endif

  /* Memory allocation for subject and allele data */
  size  = genotype->Number_of_Subjects;
  genotype->Subjects   = (Subject_Genotype *)mem_alloc(size*sizeof(Subject_Genotype),"Subject array");
  size  = genotype->Number_of_Loci;
  genotype->Locus_List = (Locus_Data *)      mem_alloc(size*sizeof(Locus_Data),      "Locus data array");
  size  = genotype->Number_of_Subjects * genotype->Number_of_Loci;
  locus                = (Locus_Code *)      mem_alloc(size*sizeof(Locus_Code),      "Genotype array for each subject");

#ifdef DEBUG
  printf("# Array size for subject = %d, for locus =  %d, for genotype = %d\n",
	 genotype->Number_of_Subjects*sizeof(Subject_Genotype),
	 genotype->Number_of_Loci    *sizeof(Locus_Data),
	 size                        *sizeof(Locus_Code));
#endif

  /* Set address for allele data */
  subject = genotype->Subjects;
  for (subject_No=0; subject_No<genotype->Number_of_Subjects; subject_No++) {
    subject->Locus = locus;
    subject++; locus = locus + genotype->Number_of_Loci;
  }

  /* Initialize allele count */
  for (locus_No=0; locus_No<genotype->Number_of_Loci; locus_No++) {
    (genotype->Locus_List + locus_No)->count[0] = 0;
    (genotype->Locus_List + locus_No)->count[1] = 0;
    (genotype->Locus_List + locus_No)->count[2] = 0;
    (genotype->Locus_List + locus_No)->genotype_count[0] = 0;
    (genotype->Locus_List + locus_No)->genotype_count[1] = 0;
    (genotype->Locus_List + locus_No)->genotype_count[2] = 0;
  }

  /* Memory allocation for work area */
  read_buffer = (char *)mem_alloc(MAX_RECORD_LENGTH*sizeof(char),"Genotype file read buffer");
  size = genotype->Number_of_Loci;
  count       = (int *)        mem_alloc(size*sizeof(int),        "Work array to count allele");
  size = genotype->Number_of_Loci * MAX_INPUT_ALLELE;
  allele_name = (Allele_Name *)mem_alloc(size*sizeof(Allele_Name),"Work array to count allele");
  for (locus_No=0; locus_No<genotype->Number_of_Loci; locus_No++) {
    *(count+locus_No) = 0;
  }

  /* Process allele data */
  rewind(File_Status->genotype_fp);
  subject = genotype->Subjects;

  while(fgets(read_buffer, MAX_RECORD_LENGTH-1, File_Status->genotype_fp) != NULL) {
    /* Delete \n */
    end = strchr(read_buffer, LINE_FEED);
    *end = STRING_TERMINATOR;

    /* Comment check and get string */
    if (!check_comment(read_buffer)) {
      ptr        = read_buffer;
      not_found  = 1;
      item_count = 0;
      locus      = subject->Locus;
      locus_No   = 0;

      while (*ptr != STRING_TERMINATOR) {

	/* If character is not delimiter, count only first character */
	if ((*ptr != DELIMITER_SPACE) && (*ptr != DELIMITER_TAB)) {
	  if (not_found) {
	    not_found = 0;
	    str = ptr;
	  }
	}
	/* If character is delimiter, save allele name */
	else {
	  if (!not_found) {
	    if (item_count == 0) {
	      *ptr = STRING_TERMINATOR;
	      if (strlen(str) > CHAR_IN_SUBJECT_NAME) {
		sprintf(error_message_buffer,"Subject name, %s, is too long. First %d characters are used.",str,CHAR_IN_SUBJECT_NAME);
		warning_message(error_message_buffer);
		*(str+CHAR_IN_SUBJECT_NAME) = STRING_TERMINATOR;
	      }
	      strcpy(subject->Name,str);
	    }
	    else if (item_count == PREFILE_PHENOTYPE) {
	      *ptr = STRING_TERMINATOR;

	      if      (strcmp(str,"0") == 0) {
		if (option->switch_p) {
		  subject->phenotype = 1;
		}
		else {
		  subject->phenotype = 0;
		}
	      }
	      else if (strcmp(str,"1") == 0) {
		if (option->switch_p) {
		  subject->phenotype = 0;
		}
		else {
		  subject->phenotype = 1;
		}
	      }
	      else {
		subject->phenotype = -1;
	      }
	      subject->QTL       = 0.0;
	    }
	    else if (item_count >= PREFILE_IGNORE_COUNT) {
	      allele_No = (item_count - PREFILE_IGNORE_COUNT) % 2;

	      *ptr = STRING_TERMINATOR;
	      /* Missing */
	      if (strcmp(str,"0") == 0) {
		strcpy(str,MISSING_DATA_NAME);
	      }
	      set_allele_code(locus_No, allele_No, str, allele_name+MAX_INPUT_ALLELE*locus_No, count+locus_No, locus+locus_No, genotype->Locus_List+locus_No);

	      if (allele_No == 1) {
		locus_No++;
	      }
	    }
	    item_count++;
	    not_found = 1;
	  }
	}

	ptr++;
      }

      /* Save last allele name */
      if (!not_found) {
	if (item_count >= PREFILE_IGNORE_COUNT) {
	  allele_No = (item_count - PREFILE_IGNORE_COUNT) % 2;

	  /* Missing */
	  if (strcmp(str,"0") == 0) {
	    strcpy(str,MISSING_DATA_NAME);
	  }
	  set_allele_code(locus_No, allele_No, str, allele_name+MAX_INPUT_ALLELE*locus_No, count+locus_No, locus+locus_No, genotype->Locus_List+locus_No);
	}
      }

      subject++;
    }
  }

  free(allele_name);
  free(count);
  free(read_buffer);

  return 0;
}

int         set_allele_code(int          locus_No,
			    int          allele_No,
			    char        *allele_name,
			    Allele_Name *list,
			    int         *count,
			    Locus_Code  *locus,
			    Locus_Data  *locus_data)
{
  int    a, not_found, err;
  allele code, mask;

  err  = 0;
  code = MISSING_DATA_CODE;

  if (strcmp(allele_name,MISSING_DATA_NAME) != 0) {
    if (*count == 0) {
      strcpy(list->Allele_Name, allele_name);
      strcpy(locus_data->Allele_Name[0], allele_name);
      locus_data->count[0] += 1;
      code    = 0;
      *count += 1;
      locus_data->number_of_allele = *count;
    }
    else {
      not_found = 1;
      for (a=0; a<*count; a++) {
	if (strcmp((list+a)->Allele_Name, allele_name) == 0) {
	  not_found = 0;
	  code = (allele)a;
	  if (a < 2) {
	    locus_data->count[a] += 1;
	  }
	  break;
	}
      }
      if (not_found) {
	if (*count < MAX_INPUT_ALLELE) {
	  if (*count == 1) {
	    strcpy(locus_data->Allele_Name[1], allele_name);
	    locus_data->count[1] += 1;
	  }
	  strcpy((list + *count)->Allele_Name, allele_name);
	  code    = *count;
	  *count += 1;
	  locus_data->number_of_allele = *count;
	  if (*count > MAX_NUMBER_OF_ALLELE) {
	    err = 1;
	    if (*count >= MAX_INPUT_ALLELE) {
	      err = 2;
	    }
	  }
	}
	else {
	  err = 3;
	}
      }
    }
  }
  else {
    locus_data->count[2] += 1;
  }

  mask = ALLELE_CODE_MASK;
  if (allele_No == 0) {
    mask = mask << ALLELE_CODE_SHIFT;
    locus->allele_code = (mask & locus->allele_code) | code;
  }
  else {
    code = code << ALLELE_CODE_SHIFT;
    locus->allele_code = (mask & locus->allele_code) | code;
  }
  /*printf("locus %3d allele No %d is %s(%d), count = %d, code = %2.2x\n",
	 locus_No,allele_No,allele_name,strlen(allele_name),
	 *count,locus->allele_code);*/

  if      (err == 1) {
    sprintf(error_message_buffer,"Number of allele, %d, at locus %d is more than 2",*count,locus_No+1);
    error_message(error_message_buffer);
  }
  else if (err == 2) {
    sprintf(error_message_buffer,"Number of allele, %d, at locus %d is too many to process",*count,locus_No);
    error_message(error_message_buffer);
  }

  return err;
}

int            check_comment(char *read_buffer)
{
  /* Local variables */
  char *ptr, *end;
  int   flag;

  flag = 1;

  /* Check # */
  end = strchr(read_buffer, COMMENT_CHAR);
  if      (end == read_buffer) {
    return flag;
  }

  /* Check \r */
  if (end == NULL) {
    end = strchr(read_buffer, CARRIAGE_RETURN);
  }

  /* Get character string */
  if (end != NULL) {
    *end = STRING_TERMINATOR;
  }

  ptr = read_buffer;
  while (*ptr != STRING_TERMINATOR) {
    if ((*ptr != DELIMITER_SPACE) && (*ptr != DELIMITER_TAB)) {
      flag = 0;
      break;
    }
    ptr++;
  }

  return flag;
}

void           free_genotype(Genotype_Data *genotype)
{
  if (genotype != NULL) {
    if (genotype->Subjects != NULL) {
      free((genotype->Subjects)->Locus);
      free(genotype->Subjects);
    }
    if (genotype->Locus_List != NULL) {
      free(genotype->Locus_List);
    }
    free(genotype);
  }

  return;
}

int            check_and_print_genotype(Genotype_Data *genotype,
					Option_Data   *option)
{
  /* Local variables */
  Subject_Genotype *subject;
  Locus_Code       *locus;
  Locus_Data       *locus_info;
  int               s, l, a, total_allele;
  int               missing_flag, missing_loci, missing_allele;
  allele            allele1, allele2, atemp;
  char              allele_name[CHAR_IN_ALLELE_NAME+1];
  char              genotype_name[3][CHAR_IN_ALLELE_NAME*2+2];
  double            frequency[2];

  /* Check multiple defined ID */
  check_Subject_ID(genotype,option);

  /* 1: Check major allele */
  locus_info = genotype->Locus_List;
  for(l=0; l<genotype->Number_of_Loci; l++) {
    if ((locus_info+l)->number_of_allele == 1) {
      strcpy((locus_info+l)->Allele_Name[1],NULL_ALLELE_NAME);
    }
#ifdef DEBUG
    printf("# Locus %5d : allele count = %5d %5d %5d total = %5d\n",l+1,
	   (locus_info+l)->count[0],
	   (locus_info+l)->count[1],
	   (locus_info+l)->count[2],
	   (locus_info+l)->count[0]+(locus_info+l)->count[1]+(locus_info+l)->count[2]);
#endif

    /* Allele 0 is not major. Change order */
    if ((locus_info+l)->count[0] < (locus_info+l)->count[1]) {
      a                        = (locus_info+l)->count[0];
      (locus_info+l)->count[0] = (locus_info+l)->count[1];
      (locus_info+l)->count[1] = a;
      strcpy(allele_name,                   (locus_info+l)->Allele_Name[0]);
      strcpy((locus_info+l)->Allele_Name[0],(locus_info+l)->Allele_Name[1]);
      strcpy((locus_info+l)->Allele_Name[1],allele_name                   );

      subject = genotype->Subjects;
      for(s=0; s<genotype->Number_of_Subjects; s++) {
	locus   = subject->Locus + l;
	allele1 = ALLELE_CODE_MASK & locus->allele_code;
	allele2 = ALLELE_CODE_MASK & (locus->allele_code >> ALLELE_CODE_SHIFT);
	if      (allele1 == 0) {
	  allele1 = 1;
	}
	else if (allele1 == 1) {
	  allele1 = 0;
	}
	if      (allele2 == 0) {
	  allele2 = 1;
	}
	else if (allele2 == 1) {
	  allele2 = 0;
	}

	locus->allele_code = (ALLELE_CODE_MASK & allele1) | ((ALLELE_CODE_MASK & allele2) << ALLELE_CODE_SHIFT);

	subject++;
      }
    }
  }

  /* 2: Sort allele ID */
  missing_flag = 0;
  subject      = genotype->Subjects;
  for(s=0; s<genotype->Number_of_Subjects; s++) {
    missing_allele = 0;
    missing_loci   = 0;

    for(l=0; l<genotype->Number_of_Loci; l++) {
      locus   = subject->Locus + l;
      allele1 = ALLELE_CODE_MASK & locus->allele_code;
      allele2 = ALLELE_CODE_MASK & (locus->allele_code >> ALLELE_CODE_SHIFT);
      if (allele1 > allele2) {
	atemp   = allele1;
	allele1 = allele2;
	allele2 = atemp;
      }

      if      (allele1 == MISSING_DATA_CODE) {
	missing_loci++;
	missing_allele += 2;
      }
      else if (allele2 == MISSING_DATA_CODE) {
	missing_loci++;
	missing_allele++;
      }
      else {
	atemp = allele1 + allele2;
	(locus_info+l)->genotype_count[atemp] += 1;
      }

      locus->allele_code = (ALLELE_CODE_MASK & allele1) | ((ALLELE_CODE_MASK & allele2) << ALLELE_CODE_SHIFT);
    }
    
    if (missing_loci != 0) {
      if (missing_flag == 0) {
	printf("\nMissing data in genotype file\n");
	printf("  subject No  subject name  number of loci  number of alleles\n");
      }
      missing_flag++;
      printf("  %10d    %10s  %14d  %17d",s+1,(genotype->Subjects + s)->Name,missing_loci,missing_allele);
      if ((option->missing_loci > 0) && (missing_loci >= option->missing_loci)) {
	printf(" : may be rejected\n");
      }
      else {
	printf("\n");
      }
    }
    subject++;
  }
  if (missing_flag > 0) {
    printf("Number of subjects which have missing data is %d\n",missing_flag);
  }

  /* 3: Calculate polymorphism parameters and statistics for HWE test */
  for(l=0; l<genotype->Number_of_Loci; l++) {
    total_allele = (locus_info+l)->count[0] + (locus_info+l)->count[1];
    (locus_info+l)->Allele_Frequency = (double)(locus_info+l)->count[0]/(double)total_allele; 
    frequency[0] =       (locus_info+l)->Allele_Frequency;
    frequency[1] = 1.0 - (locus_info+l)->Allele_Frequency;
    (locus_info+l)->PIC            = calc_PIC((locus_info+l)->number_of_allele, frequency);
    (locus_info+l)->heterozygosity = calc_Heterozygosity((locus_info+l)->number_of_allele, total_allele, frequency, &((locus_info+l)->hetero_variance));
    (locus_info+l)->HWE_check      = calc_HWE_statistics((locus_info+l)->genotype_count);
    (locus_info+l)->p_value        = calc_p_value_by_chi_square(1, (locus_info+l)->HWE_check);
#ifdef DEBUG
    printf("# Locus %5d : genotype count = %5d %5d %5d total = %5d\n",l+1,
	   (locus_info+l)->genotype_count[0],
	   (locus_info+l)->genotype_count[1],
	   (locus_info+l)->genotype_count[2],
	   (locus_info+l)->genotype_count[0]+(locus_info+l)->genotype_count[1]+(locus_info+l)->genotype_count[2]);
#endif
  }

  printf("\nNumber of subjects = %d, number of loci = %d in genotype file\n\n",
	 genotype->Number_of_Subjects,genotype->Number_of_Loci);

  printf("Alleles for each locus\n");
  printf("               Major allele          Minor allele        Missing\n");
  printf("  locus    name count frequency  name count frequency  name count   total\n");
  for(l=0; l<genotype->Number_of_Loci; l++) {
    total_allele = (locus_info+l)->count[0] + (locus_info+l)->count[1] + (locus_info+l)->count[2];
    printf("%7d   %5s %5d %9.4f %5s %5d %9.4f %5s %5d %7d\n",l+1,
	   (locus_info+l)->Allele_Name[0],(locus_info+l)->count[0],    (locus_info+l)->Allele_Frequency,
	   (locus_info+l)->Allele_Name[1],(locus_info+l)->count[1],1.0-(locus_info+l)->Allele_Frequency,
	   MISSING_DATA_NAME,             (locus_info+l)->count[2],total_allele);
  }

  printf("Polymorphism parameters for each locus\n");
  printf("  locus         PIC  heterozygosity  variance  statistics for HWE    p-value\n");
  for(l=0; l<genotype->Number_of_Loci; l++) {
    printf("%7d   %9.4f %15.4f %9.2e %19.4f  %9.2e\n",l+1,
	   (locus_info+l)->PIC,
	   (locus_info+l)->heterozygosity,
	   (locus_info+l)->hetero_variance,
	   (locus_info+l)->HWE_check,
	   (locus_info+l)->p_value);
  }

  printf("Genotype counts for each locus\n");
  printf("  locus    genotype count    genotype count    genotype count  total\n");
  for(l=0; l<genotype->Number_of_Loci; l++) {
    strcpy(genotype_name[0],(locus_info+l)->Allele_Name[0]);
    strcat(genotype_name[0],"/");
    strcat(genotype_name[0],(locus_info+l)->Allele_Name[0]);
    strcpy(genotype_name[1],(locus_info+l)->Allele_Name[0]);
    strcat(genotype_name[1],"/");
    strcat(genotype_name[1],(locus_info+l)->Allele_Name[1]);
    strcpy(genotype_name[2],(locus_info+l)->Allele_Name[1]);
    strcat(genotype_name[2],"/");
    strcat(genotype_name[2],(locus_info+l)->Allele_Name[1]);
    if ((locus_info+l)->number_of_allele == 1) {
      printf("%7d %11s %5d                                      %5d\n",l+1,
	     genotype_name[0],(locus_info+l)->genotype_count[0],
	     (locus_info+l)->genotype_count[0]+(locus_info+l)->genotype_count[1]+(locus_info+l)->genotype_count[2]);
    }
    else {
      printf("%7d %11s %5d %11s %5d %11s %5d  %5d\n",l+1,
	     genotype_name[0],(locus_info+l)->genotype_count[0],
	     genotype_name[1],(locus_info+l)->genotype_count[1],
	     genotype_name[2],(locus_info+l)->genotype_count[2],
	     (locus_info+l)->genotype_count[0]+(locus_info+l)->genotype_count[1]+(locus_info+l)->genotype_count[2]);
    }
  }
  

  if (option->message) {
    printf("\nGenotype for each subject\n");
    printf("  subject name   genotype\n");
    subject = genotype->Subjects;
    for(s=0; s<genotype->Number_of_Subjects; s++) {
      printf("  %12s  ",subject->Name);
      for(l=0; l<genotype->Number_of_Loci; l++) {
	locus   = subject->Locus + l;
	allele1 = ALLELE_CODE_MASK & locus->allele_code;
	allele2 = ALLELE_CODE_MASK & (locus->allele_code >> ALLELE_CODE_SHIFT);
	if (allele1 != MISSING_DATA_CODE) {
	  printf(" %s",(locus_info+l)->Allele_Name[allele1]);
	}
	else {
	  printf(" %s",MISSING_DATA_NAME);
	}
	if (allele2 != MISSING_DATA_CODE) {
	  printf("/%s",(locus_info+l)->Allele_Name[allele2]);
	}
	else {
	  printf("/%s",MISSING_DATA_NAME);
	}
      }
      printf("\n");
      subject++;
    }
  }

  return 0;
}

int            check_Subject_ID(Genotype_Data *genotype,
				Option_Data   *option)
{
  /* Local variables */
  int           s, t, RC;

  RC = 0;
  if (option->phenotype_format == 0) {
    for(s=0; s<genotype->Number_of_Subjects-1; s++) {
      for(t=s+1; t<genotype->Number_of_Subjects; t++) {
	if (strcmp((genotype->Subjects + s)->Name,(genotype->Subjects + t)->Name) == 0) {
	  sprintf(error_message_buffer,"Subject ID '%s' of subject No %d is same as ID of subject No %d",(genotype->Subjects + t)->Name,t+1,s+1);
	  error_message(error_message_buffer);
	}
      }
    }
  }
  if (error_check()) {
    RC = 1;
  }

  return RC;
}
