#include "haplotypetest.h"

/* Utility function */
extern void          *mem_alloc(size_t, const char *);
extern void           hpsort_reverse_index_of_double_array(int, double *, unsigned int *);
extern void           hpsort_index_of_double_array(int, double *, unsigned int *);

/* Error handler */
extern int            error_check();
extern int            error_message(char *);
extern int            warning_message(char *);

/* Problem definition for EM */
extern Estimate_Set  *setup_EM(Genotype_Data *, Estimate_Set *, int);
extern Estimate_Set  *combination(Estimate_Set *);
extern int            all_haplotype(Estimate_Set *);
extern int            set_diplotype_class(Estimate_Set *);
extern int            trait_inheritance_model(Estimate_Set *, int, int);
extern int            no_association_model(Estimate_Set *);
extern void           free_estimate_set(Estimate_Set *);
extern Estimate_Set  *alloc_estimate_set();

/* Frequency initialization */
extern int            Init_frequency(Estimate_Set *, double *, int);

/* EM engine */
extern double         EM_iteration_with_penetrance(Estimate_Set *, Option_Data *, double **, double **, double *, int);
extern int            count_nonzero_haplotype(int, double *, haplotype *);
extern int            estimated_allele_frequency(Estimate_Set *, int, haplotype *, double *, double *);

/* Functions for no LD condition */
extern double         noLD_likelihood_with_penetrance(Estimate_Set *, double *, int);
extern int            noLD_frequency(Estimate_Set *, double *, double *);

/* Result print */
extern int            print_allele_table(Estimate_Set *, Genotype_Data *, double *);
extern int            print_haplotype_frequency_with_parameters(Estimate_Set *, Estimate_Set *, double *, Genotype_Data *);
extern int            print_penetrances(Estimate_Set *, double [][2], double, int);
extern int            print_diplotype_profile_with_penetrance(Estimate_Set *, Genotype_Data *);
extern int            print_p_value_of_LRC(int, double);
extern int            print_contingency_table_for_penetrance(Estimate_Set *, double, double [2], double, int);
extern double         calc_p_value_by_chi_square(int, double);
extern int            print_test_of_haplotype_set_with_penetrance(Estimate_Set *, Genotype_Data *, Haplotype_Set *, unsigned int *, double, int);
extern int            sprint_haplotype_with_mask(char *, haplotype, haplotype, Estimate_Set *, Genotype_Data *);

/* Standard error calculation by non-parametric bootstrap */
extern int            bootstrap_with_penetrance(Genotype_Data *, Estimate_Set *, Option_Data *);

/* Odds ratio calculation for case-control study */
extern double         odds_ratio_by_penetrance(Estimate_Set *);

/* P value calculation by permutation test */
extern double         permutation_with_penetrance(Genotype_Data *, Estimate_Set *, Option_Data *, double, int);

/* Confidence interval calculation */
extern double         Total_likelihood_with_penetrance(Estimate_Set *, double *);
extern int            Confidence_interval_for_penetrance(Estimate_Set *, Option_Data *, double [][2], double ());
extern int            Confidence_interval_for_odds_ratio(Estimate_Set *, Option_Data *, double [2], double ());

/* Outlier detection */
extern double         Subject_likelihood_with_penetrance(Estimate_Set *, double *, double *);
extern int            Outlier_by_genotype_set(Estimate_Set *, Option_Data *, Genotype_Data *, double *, double ());

#ifdef DEBUG
extern int            dump_estimate_set(Estimate_Set *);
#endif

/* Define functions */
int            estimate_haplotype_and_penetrance(Genotype_Data *, Estimate_Set *, Haplotype_Set *, Option_Data *, Window_Result *);

int            estimate_haplotype_and_penetrance(Genotype_Data *genotype,
						 Estimate_Set  *table,
						 Haplotype_Set *set_of_haplotype,
						 Option_Data   *option,
						 Window_Result *test_result)
{
  /* Local variables */
  int            size, l, nonzero, init_save;
  int            count_max, index_max, count0_max, index0_max;
  int            haplo_set, o_flag1, o_flag2, null_hypothesis, model_type, o_trait;
  haplotype      h, mask;
  double         criteria;
  double         LRC_conv = 2.0 * log(10.0);
  int            save_permutation;
  double        *p_value;
  unsigned int  *p_order;

  /* Frequencies */
  double        *f_old, *f_new, *f_subject, *f_noLD, *f_wk, *f_0, *f_tmp;
  double        *f_new_r, *f_0_r;
  double        *f_allele_save, *f_allele;
  haplotype     *order, *order_tmp;

  /* Parameter for no association for target */
  Estimate_Set  *table_for_0;

  /* Parameter for maximum likelihood estimation */
  double         likelihood, likelihood0, likelihood0t, LRC_max;
  double         likelihood_max, likelihood0_max;

  double    (*likelihood_func_sub)(Estimate_Set *, double *, double *) = Subject_likelihood_with_penetrance;
  double    (*likelihood_function)(Estimate_Set *, double *)           = Total_likelihood_with_penetrance;

  /* Output control */
  if (option->msg_flag) {
    if (option->haplotype_set) {
      o_flag1 = option->message;
    }
    else {
      o_flag1 = 1;
    }
  }
  else {
    o_flag1 = 0;
  }
  if (o_flag1) {
    o_flag2 = option->maximize_f;
  }
  else {
    o_flag2 = 0;
  }

  /* Set up tables for haplotype inferance */
  table = setup_EM(genotype, table, option->missing_loci);
  if (option->msg_flag) {
    printf("\nNumber of effective subjects to be used             =  %d\n",genotype->Number_of_Effective_Subjects);
    printf("Number of genotype for EM calculation               =  %d\n",table->number_of_set);
    if (genotype->Number_of_Unknown > 0) {
      printf("Number of phenotype unknown subjects to be used     =  %d\n",genotype->Number_of_Unknown);
    }
  }

  /* Make combination table */
  table = combination(table);
  all_haplotype(table);
  if (option->msg_flag) {
    printf("Number of possible haplotypes in the problem        =  %d\n",table->number_of_haplo);
  }

  /* Allocate array for p values */
  size         = set_of_haplotype->number_of_subset;
  p_value      = (double       *)mem_alloc(size*sizeof(double),"Array for p values");
  p_order      = (unsigned int *)mem_alloc(size*sizeof(unsigned int),"Array for sorting p values");

  /* Allocate frequencies and expectation */
  size         = table->number_of_haplo;
  table->order = (haplotype *)mem_alloc(size*sizeof(haplotype),"Array for sort frequency");
  f_old        = (double *)mem_alloc(size*sizeof(double),"Frequency array for old step");
  f_new        = (double *)mem_alloc(size*sizeof(double),"Frequency array for new step");
  f_0          = (double *)mem_alloc(size*sizeof(double),"Frequency array for no association");
  f_noLD       = (double *)mem_alloc(size*sizeof(double),"Frequency array for no LD");
  f_subject    = (double *)mem_alloc(size*sizeof(double),"Frequency array for each subject");

  /* Allocate table for no association */
  table_for_0              = alloc_estimate_set();
  *table_for_0             = *table;
  table_for_0->geno_info   = (Genotype_Pack *)mem_alloc(table->number_of_set*sizeof(Genotype_Pack),"Packed genotype structure table");
  table_for_0->target_mask = 0;
  table_for_0->target      = 0;
  table_for_0->model_type  = 0;
  for (l=0; l<table_for_0->number_of_set; l++) {
    *(table_for_0->geno_info + l) = *(table->geno_info + l);
  }
  f_allele_save  = (double *)mem_alloc(table->select_loci*sizeof(double),"Allele frequency for selected loci");
  for (l=0; l<table->select_loci; l++) {
    *(f_allele_save + l)   = *(table->f_allele + l);
  }

  if (option->maximize > 0) {
    if (option->msg_flag) {
      printf("Restart count to search maximum LRC                 =  %d\n",option->maximize);
    }

    /* Allocate frequencies */
    f_new_r    = (double *)mem_alloc(size*sizeof(double),"Frequency array for new step");
    f_0_r      = (double *)mem_alloc(size*sizeof(double),"Frequency array for no association");
    f_wk       = (double *)mem_alloc(size*sizeof(double),"Frequency array for no LD");
    f_allele   = (double *)mem_alloc(table->select_loci*sizeof(double),"Allele frequency for selected loci");
    order      = (haplotype *)mem_alloc(size*sizeof(haplotype),"Array for sort frequency");
  }

  /* Loop for mode of inheritance */
  for (model_type=0; model_type<option->number_of_model; model_type++) {

    /* Loop for haplotype set */
    null_hypothesis   = 1;
    if (option->msg_flag) {
      o_trait           = 1;
    }
    else {
      o_trait           = 0;
    }
    table->model_type = option->model_type[model_type];
    for (haplo_set=0; haplo_set<set_of_haplotype->number_of_subset; haplo_set++) {
      table->target      = (set_of_haplotype->subset + haplo_set)->target;
      table->target_mask = (set_of_haplotype->subset + haplo_set)->mask;

      /* Set diplotype flags */
      set_diplotype_class(table);
      table->df_for_test = trait_inheritance_model(table, table->model_type, o_trait);
      o_trait            = 0;

      if (o_flag1 && option->maximize > 0) {
	printf("Information for first EM iteration of given initialization method as follows\n");
      }

      /* Initialize probability */
      Init_frequency(table, f_new, option->init_type);
      if (option->init_type == 2 || null_hypothesis) { /* Random */
	for (h=0; h<table->number_of_haplo; h++) {
	  *(f_0 + h) = *(f_new + h);
	}
      }

      /* EM iteration */
      likelihood   = EM_iteration_with_penetrance(table, option, &f_old, &f_new, f_subject, o_flag1);
      table->f     = f_new;

      if (option->init_type == 2 || null_hypothesis) { /* Random */
	null_hypothesis = 0;

	/* Sort haplotype frequency */
	hpsort_reverse_index_of_double_array(table->number_of_haplo, table->f, table->order);
	table->nonzero = count_nonzero_haplotype(table->number_of_haplo, table->f, table->order);

	/* Final allele frequency */
	for (l=0; l<table->select_loci; l++) {
	  *(table->f_allele + l) = 0.0;
	}
	estimated_allele_frequency(table, table->nonzero, table->order, table->f, table->f_allele);
 
	/* No LD likelihood */
	noLD_frequency(table, table->f_allele, f_noLD);
	likelihood0  = noLD_likelihood_with_penetrance(table, f_noLD, o_flag1);

	/* No association likelihood */
#ifdef DEBUG_PENETRANCE
	dump_estimate_set(table);
#endif

	set_diplotype_class(table_for_0);
	no_association_model(table_for_0);
	likelihood0  = EM_iteration_with_penetrance(table_for_0, option, &f_old, &f_0, f_subject, 0);

#ifdef DEBUG_PENETRANCE
	printf("# for no association\n");
	dump_estimate_set(table_for_0);
#endif

	table_for_0->f = f_0;
      }

      LRC_max         = likelihood - likelihood0;
      likelihood_max  = likelihood;
      likelihood0_max = likelihood0;

      if (o_flag1) {
	printf("log10 likelihood assuming no association            = %13.6e\n",likelihood0);
	printf("Final log10 Likelihood                              = %13.6e\n",likelihood);
	printf("Likelihood Ratio Criteria (LRC)                     = %13.6e\n",LRC_max);
      }

      /* Maximum LRC search */
      if (option->maximize > 0) {
	if (o_flag2) {
	  /* Output haplotype frequencies and diplotype configuration */
	  print_allele_table(table, genotype, f_allele_save);
	  print_haplotype_frequency_with_parameters(table, table_for_0, f_noLD, genotype);
	  print_penetrances(table, (set_of_haplotype->subset + haplo_set)->interval, option->confidence, 0);
	}

	if (o_flag1) {
	  printf("\nInformation of restart to search maximum LRC\n");
	}
	criteria          = log10(option->criteria)*2.0*fabs(likelihood_max);
	count_max         = 0;
	index_max         = 0;
	count0_max        = 0;
	index0_max        = 0;
	init_save         = option->init_type;
	option->init_type = 2; /* Random distribution */

	/* Loop for LRC search */
	for (l=1; l<=option->maximize; l++){
	  if (o_flag2) {
	    printf("\nRestart No. %d\n",l);
	  }

	  /* EM iteration */
	  Init_frequency(table, f_new_r, option->init_type);
	  for (h=0; h<table->number_of_haplo; h++) {
	    *(f_0_r + h) = *(f_new_r + h);
	  }
	  likelihood   = EM_iteration_with_penetrance(table, option, &f_old, &f_new_r, f_subject, o_flag2);

	  /* Sort haplotype frequency */
	  hpsort_reverse_index_of_double_array(table->number_of_haplo, f_new_r, order);
	  nonzero = count_nonzero_haplotype(table->number_of_haplo, f_new_r, order);

	  /* Final allele frequency */
	  estimated_allele_frequency(table, nonzero, order, f_new_r, f_allele);

	  /* No LD likelihood */
	  noLD_frequency(table, f_allele, f_wk);
	  likelihood0t = noLD_likelihood_with_penetrance(table, f_wk, o_flag2);
	  likelihood0t = EM_iteration_with_penetrance(table_for_0, option, &f_old, &f_0_r, f_subject, 0);

	  if (o_flag2) {
	    printf("log10 likelihood assuming no association            = %13.6e\n",likelihood0t);
	    printf("Final log10 Likelihood                              = %13.6e\n",likelihood);
	  }

	  /* Check Likelihood */
	  if      (fabs(likelihood - likelihood_max) < criteria) {
	    count_max++;
	  }
	  else if (likelihood - likelihood_max > criteria) {
	    count_max       = 0;
	    index_max       = l;
	    likelihood_max  = likelihood;

	    f_tmp           = table->f;
	    table->f        = f_new_r;
	    f_new_r         = f_tmp;
	    f_new           = table->f;
	    f_tmp           = f_noLD;
	    f_noLD          = f_wk;
	    f_wk            = f_tmp;
	    f_tmp           = table->f_allele;
	    table->f_allele = f_allele;
	    f_allele        = f_tmp;
	    order_tmp       = table->order;
	    table->order    = order;
	    order           = order_tmp;
	    table->nonzero  = nonzero;

	    if (o_flag1) {
	      printf("Restart No. %5d : Likelihood for alternative      = %13.6e\n",l,likelihood);
	    }
	    if (o_flag2) {
	      /* Output haplotype frequencies */
	      print_allele_table(table, genotype, f_allele);
	      print_haplotype_frequency_with_parameters(table, table_for_0, f_noLD, genotype);
              print_penetrances(table, (set_of_haplotype->subset + haplo_set)->interval, option->confidence, 0);
	    }
	  }

	  if      (fabs(likelihood0t - likelihood0_max) < criteria) {
	    count0_max++;
	  }
	  else if (likelihood0t - likelihood0_max > criteria) {
	    count0_max      = 0;
	    index0_max      = l;
	    likelihood0_max = likelihood0t;

	    f_tmp           = table_for_0->f;
	    table_for_0->f  = f_0_r;
	    f_0_r           = f_tmp;
	    f_0             = table_for_0->f;

	    if (o_flag1) {
	      printf("Restart No. %5d : Likelihood for null             = %13.6e\n",l,likelihood0t);
	    }
	    if (o_flag2) {
	      /* Output haplotype frequencies */
	      print_allele_table(table, genotype, f_allele);
	      print_haplotype_frequency_with_parameters(table, table_for_0, f_noLD, genotype);
              print_penetrances(table, (set_of_haplotype->subset + haplo_set)->interval, option->confidence, 0);
	    }
	  }
	}
	LRC_max = likelihood_max - likelihood0_max;
	if (o_flag1) {
	  printf("  Final Likelihood for alternative hypothesis       = %13.6e at %d\n",likelihood_max,index_max);
	  printf("  Same Likelihood came out %d times\n",count_max);
	  printf("  Final Likelihood for null hypothesis              = %13.6e at %d\n",likelihood0_max,index0_max);
	  printf("  Same Likelihood came out %d times\n",count0_max);
	  printf("  Final LRC                                         = %13.6e\n",LRC_max);
	}
    
	option->init_type = init_save;
      }

      if (LRC_max < 0) {
	(set_of_haplotype->subset + haplo_set)->statistic = 0.0;
      }
      else {
	(set_of_haplotype->subset + haplo_set)->statistic = LRC_max * LRC_conv;
      }
      (set_of_haplotype->subset + haplo_set)->p_value   = calc_p_value_by_chi_square(table->df_for_test, (set_of_haplotype->subset + haplo_set)->statistic);
      (set_of_haplotype->subset + haplo_set)->result[0]  = table->p_plus[0];
      (set_of_haplotype->subset + haplo_set)->result[1]  = table->p_plus[1];
      (set_of_haplotype->subset + haplo_set)->result[2]  = table->p_plus[2];
      (set_of_haplotype->subset + haplo_set)->odds_ratio = odds_ratio_by_penetrance(table);
      *(p_value + haplo_set)                             = (set_of_haplotype->subset + haplo_set)->p_value;

      if (option->conf_flag) {
	Confidence_interval_for_penetrance(table, option, (set_of_haplotype->subset + haplo_set)->interval, likelihood_function);
	Confidence_interval_for_odds_ratio(table, option, (set_of_haplotype->subset + haplo_set)->odds_interval, likelihood_function);
      }

      save_permutation = option->permutation;
      if (option->permutation < 0) {
	if ((set_of_haplotype->subset + haplo_set)->p_value > 0.0) {
	  option->permutation = 1;
	  while ((double)option->permutation * (set_of_haplotype->subset + haplo_set)->p_value < 10.0) {
	    option->permutation = option->permutation * 10;
	  }
	}
	else {
	  option->permutation = 0;
	  (set_of_haplotype->subset + haplo_set)->p_value_p = 0.0;
	}
      }

      if (option->permutation) {
	(set_of_haplotype->subset + haplo_set)->p_value_p = permutation_with_penetrance(genotype, table, option, LRC_max, o_flag1);
      }
      option->permutation = save_permutation;

      /* Output haplotype frequencies and diplotype configuration */
      if (o_flag1) {
	print_p_value_of_LRC(table->df_for_test, LRC_max);
	print_contingency_table_for_penetrance(table, (set_of_haplotype->subset + haplo_set)->odds_ratio, (set_of_haplotype->subset + haplo_set)->odds_interval, option->confidence, option->conf_flag);
	print_penetrances(table, (set_of_haplotype->subset + haplo_set)->interval, option->confidence, option->conf_flag);
	print_allele_table(table, genotype, f_allele_save);
	print_haplotype_frequency_with_parameters(table, table_for_0, f_noLD, genotype);
	if (option->outlier) {
	  Outlier_by_genotype_set(table, option, genotype, table->f, likelihood_func_sub);
	}
	if (option->diplo_out) {
	  print_diplotype_profile_with_penetrance(table, genotype);
	  print_diplotype_profile(table, genotype);
	}
      }
    }

    /* Print result of test of haplotype set */
    if (option->haplotype_set ) {
      hpsort_index_of_double_array(set_of_haplotype->number_of_subset, p_value, p_order);
      if (test_result != NULL) {
	(test_result + model_type)->statistic = (set_of_haplotype->subset + *p_order)->statistic;
	(test_result + model_type)->p_value   = (set_of_haplotype->subset + *p_order)->p_value;
	for (l=0; l<3; l++) {
	  (test_result + model_type)->result[l]      = (set_of_haplotype->subset + *p_order)->result[l];
	  (test_result + model_type)->interval[l][0] = (set_of_haplotype->subset + *p_order)->interval[l][0];
	  (test_result + model_type)->interval[l][1] = (set_of_haplotype->subset + *p_order)->interval[l][1];
	}
	(test_result + model_type)->odds_ratio       = (set_of_haplotype->subset + *p_order)->odds_ratio;
	(test_result + model_type)->odds_interval[0] = (set_of_haplotype->subset + *p_order)->odds_interval[0];
	(test_result + model_type)->odds_interval[1] = (set_of_haplotype->subset + *p_order)->odds_interval[1];
	sprint_haplotype_with_mask((test_result + model_type)->haplotype_name,(set_of_haplotype->subset + *p_order)->target, (set_of_haplotype->subset + *p_order)->mask, table, genotype);
      }
      if (option->msg_flag) {
	print_test_of_haplotype_set_with_penetrance(table, genotype, set_of_haplotype, p_order, option->confidence, option->conf_flag);
      }
    }

    /* Bootstrap calculation start */
    if (option->bootstrap > 0) {
      bootstrap_with_penetrance(genotype, table, option);
    }
    /* Bootstrap calculation end */
  }

  /* Free allocated area */
  free(table_for_0->class);
  free(table_for_0->geno_info);
  free(table_for_0->f);
  free(table_for_0);
  if (option->maximize > 0) {
    free(order);
    free(f_allele);
    free(f_wk);
    free(f_0_r);
    free(f_new_r);
  }
  free(f_allele_save);
  free(f_subject);
  free(f_noLD);
  free(f_old);
  free(p_order);
  free(p_value);

  return 0;
}
